//This file is part of the LCongruenceClass v0.1 library written in -*- C++ -*-

//LCongruenceClass v0.1: a simple congruence class implementation
//Copyright (C) Luca Saiu, 1999

//This library is free software; you can redistribute it and/or modify it
//under the terms of the GNU Library General Public License version 2
//as published by the Free Software Foundation.

//This library is distributed in the hope that it will be useful, but
//WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//Library General Public License for more details.

//You should have received a copy of the GNU Library General Public
//License along with this library; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


//Luca Saiu
//------------------------------------------
//e-mail:      luca78@dada.it
//paper mail:  Via Camporcioni, 15
//             51019 Ponte Buggianese - Pistoia
//             Italy
//telephone:   0572-635477 Italy


#ifndef _LCONGRUENCECLASS_H_
#define _LCONGRUENCECLASS_H_

#include <iostream.h>

template <long m> class LCongruenceClass{
protected:
  long a;                                       // Stored value

  template <long n> friend LCongruenceClass<n>; // To be able to read private
                                                // and protected members of
                                                // LCongruenceClasses even
                                                // when modules are different

  long FindBest() const;                        // Finds the best representant
public:
  // Constructors (there is no need for destructors)
  LCongruenceClass();                           // Default constructor
  LCongruenceClass(const long x);               // Conversion constructor
  LCongruenceClass(const LCongruenceClass<m> &x);//Copy constructor

  // Conversion operator
  operator long() const;

  // Assignment operators
  template <long n> const LCongruenceClass<m>& operator=(const LCongruenceClass<n> x);
  const LCongruenceClass<m>& operator=(const long x);

  // Mathematical unary plus and minus operators
  const LCongruenceClass<m> operator+() const;
  const LCongruenceClass<m> operator-() const;

  // Mathematical increment/decrement operators
  const LCongruenceClass<m>& operator++();
  const LCongruenceClass<m>& operator++(int);
  const LCongruenceClass<m>& operator--();
  const LCongruenceClass<m>& operator--(int);

  // Mathematical binary operators
  template <long n> const LCongruenceClass<m> operator+(const LCongruenceClass<n> x) const;
  template <long n> const LCongruenceClass<m> operator-(const LCongruenceClass<n> x) const;
  template <long n> const LCongruenceClass<m> operator*(const LCongruenceClass<n> x) const;

  // Mathematical combined operators
  template <long n> const LCongruenceClass<m>& operator+=(const LCongruenceClass<n> x);
  template <long n> const LCongruenceClass<m>& operator-=(const LCongruenceClass<n> x);
  template <long n> const LCongruenceClass<m>& operator*=(const LCongruenceClass<n> x);

  // Method to find the reverse element, if it exists
  const LCongruenceClass<m> Reverse() const;

  // Compare operators
  template <long n> bool operator==(const LCongruenceClass<n> x) const;
  template <long n> bool operator!=(const LCongruenceClass<n> x) const;

  // Stream functions
  friend istream& operator>> <>(istream &s,LCongruenceClass<m> &x);
  // There is no need for an ostream function, because the default conversion
  // into long does the same thing
};

// Implementation starts here.
///////////////////////////////////////////////////////////////////////////

// There are not many comments but _everything_ here is standard for a
// mathematical class.

#include "Lgen/Lgen.h"
#include "Lstring/Lstring.h"
#include <stdio.h>
#include <iostream.h>

template <long m> LCongruenceClass<m>::LCongruenceClass<m>(){
  a=0; // I prefer to have a known default value
}

template <long m> LCongruenceClass<m>::LCongruenceClass(const LCongruenceClass<m>
                                                    &x){
  a=x.a;
}

template <long m> LCongruenceClass<m>::LCongruenceClass<m>(const long x){
  a=x;
  a=FindBest();
}

template <long m> long LCongruenceClass<m>::FindBest() const{
  long r=a%m;

  if(r>=0)      // a % m might be negative...
    return r;
  else
    return r+m; // ...it's better to always use the minimum positive number.
}

template <long m> LCongruenceClass<m>::operator long() const{
  return FindBest();
}

template <long m> template <long n> const LCongruenceClass<m>&
              LCongruenceClass<m>::operator=(const LCongruenceClass<n> x){
  if(m!=n)
    fatal("LCongruenceClass: binary operator with incompatible modules");
  a=x.a;
  return *this;
}

template <long m> const LCongruenceClass<m>& LCongruenceClass<m>::operator=
                        (const long x){
  a=x;
  a=FindBest();
  return *this;
}

template <long m> const LCongruenceClass<m> LCongruenceClass<m>::operator+() const{
  return *this;
}

template <long m> const LCongruenceClass<m> LCongruenceClass<m>::operator-() const{
  LCongruenceClass r=*this;
  r.a*=-1;
  r=r.FindBest();
  return r;
}

template <long m> template <long n> const LCongruenceClass<m> LCongruenceClass<m>::operator+(const LCongruenceClass<n> x) const{
  LCongruenceClass r;

  if(m!=n)
    fatal("LCongruenceClass: binary operator with incompatible modules");
  r.a=a+x.a;
  r=r.FindBest(); // To avoid overflows
  return r;
}

template <long m> template <long n> const LCongruenceClass<m> LCongruenceClass<m>::operator-(const LCongruenceClass<n> x) const{
  LCongruenceClass r;

  if(m!=n)
    fatal("LCongruenceClass: binary operator with incompatible modules");
  r.a=a-x.a;
  r=r.FindBest(); // To avoid overflows
  return r;
}

template <long m> template <long n> const LCongruenceClass<m> LCongruenceClass<m>::operator*(const LCongruenceClass<n> x) const{
  LCongruenceClass r;

  if(m!=n)
    fatal("LCongruenceClass: binary operator with incompatible modules");
  r.a=a*x.a;
  r=r.FindBest(); // To avoid overflows
  return r;
}

template <long m> template <long n> const LCongruenceClass<m>& LCongruenceClass<m>::operator+=(const LCongruenceClass<n> x){
  if(m!=n)
    fatal("LCongruenceClass: binary operator with incompatible modules");
  a+=x.a;
  a=FindBest(); // To avoid overflows
  return *this;


template <long m> template <long n> const LCongruenceClass<m>& LCongruenceClass<m>::operator-=(const LCongruenceClass<n> x){
  if(m!=n)
    fatal("LCongruenceClass: binary operator with incompatible modules");
  a+=x.a;
  a=FindBest(); // To avoid overflows
  return *this;
}

template <long m> template <long n> const LCongruenceClass<m>& LCongruenceClass<m>::operator*=(const LCongruenceClass<n> x){
  if(m!=n)
    fatal("LCongruenceClass: binary operator with incompatible modules");
  a*=x.a;
  a=FindBest(); // To avoid overflows
  return *this;
}

template <long m> const LCongruenceClass<m>& LCongruenceClass<m>::operator++(){
  a++;
  a=FindBest(); // To avoid overflows
  return *this;
}

template <long m> const LCongruenceClass<m>& LCongruenceClass<m>::operator--(){
  a--;
  a=FindBest(); // To avoid overflows
  return *this;
}

template <long m> const LCongruenceClass<m>& LCongruenceClass<m>::operator++(int){
  return operator++();
}

template <long m> const LCongruenceClass<m>& LCongruenceClass<m>::operator--(int){
  return operator--();
}

template <long m> template <long n> bool LCongruenceClass<m>::operator==(const LCongruenceClass<n> x) const{
  if(m!=n)
    fatal("LCongruenceClass: binary operator with incompatible modules");
  return (a==x.a); // Values are always stored as the minimum natural number
                   // so I can directly compare them
}

template <long m> template <long n> bool LCongruenceClass<m>::operator!=(const LCongruenceClass<n> x) const{
  if(m!=n)
    fatal("LCongruenceClass: binary operator with incompatible modules");
  return (a!=x.a); // Values are always stored as the minimum natural number
                   // so I can directly compare them
}

template <long m> const LCongruenceClass<m> LCongruenceClass<m>::Reverse() const{
  long i;

  for(i=1;i<m;i++)
    if(((a*i)%m)==1)   // If ai = 1 (mod m)...
       return i;       // ...I have found the reverse

  // The reverse does not exist. To Do: should I compute (a,m) instead
  // and test before starting? I think not right now, I will nearly always
  // work with fields, so this solution is faster...
  fatal(Lstring(FindBest())+"x = 1 (mod "+Lstring(m)+") has no solution");
  return 0;            // Just to avoid a compiler warning
}

template <long m> istream& operator>>(istream &s, LCongruenceClass<m> &x){
  return (s>>((long&)x));
}

#endif //#ifndef _LCONGRUENCECLASS_H_


